package	jp.co.sds_corp.transfer.output;

import java.util.Collections;
import java.util.EventObject;
import java.util.HashMap;
import java.util.Map;

import jp.co.sds_corp.transfer.IDataKey;
import jp.co.sds_corp.transfer.IReceiveListener;
import jp.co.sds_corp.transfer.IReceiver;
import jp.co.sds_corp.transfer.ReceiveEvent;
import jp.co.sds_corp.transfer.ReceiverSupport;
import jp.co.sds_corp.transfer.StatusChangeEvent;
import jp.co.sds_corp.util.common.BlockQueue;
import jp.co.sds_corp.util.config.IConfiguration;
import jp.co.sds_corp.util.log.CmnLogFactory;
import jp.co.sds_corp.util.log.CmnLogger;

/**
* データ受信イベントを送信します。
このAdpaterは、同一VM内のEventSourceにのみデータ受信イベントを送信します。
<PRE>
&lt;Output name="Event" class="jp.co.sds_corp.transfer.output.EventAdapter"&gt;
	&lt;name&gt;NAME&lt;/name&gt;
&lt;/Output&gt;
</PRE>
name要素によって、出力元を識別します。
*/
public class EventAdapter extends AbstractOutput implements IReceiver
{
	static final CmnLogger logger = CmnLogFactory.getLogger();
	static Map<String,EventAdapter> map = Collections.synchronizedMap( new HashMap<String,EventAdapter>());
	String	name;
	BlockQueue<EventObject>	queue = new BlockQueue<EventObject>();
	ServiceThread thr;
	ReceiverSupport support = new ReceiverSupport();

	/**
	* 指定された初期情報で初期化します。
	* @param	config	初期情報
	*/
	public void init( IConfiguration config ) throws Exception {
		super.init( config );
		name = config.get( "name" );

		Object obj = map.get( name );
		if ( obj != null ) {
			throw( new Exception( "name already registered. "
								+ "[" + name + ":" + obj + "]" ));
		}
		map.put( name, this );
	}

	/**
	* このサービスを開始します。
	*/
	public void start() {
		try {
			thr = new ServiceThread();
			thr.setName( getName() + ":" + name );
			thr.start();
			setInterrupt( false );
		} catch ( Exception e ) {
			logger.error( this, "error", e );
		}
	}

	/**
	* このサービスを停止します。
	*/
	public void stop() {
		setInterrupt( true );
		Thread.currentThread().interrupt();
		thr.interrupt();
		try {
			thr.join();
		} catch ( Exception e ) {
		}
		logger.debug( this, "stopped." );
	}

	/**
	* 指定された名前のEventAdapterを生成します。
	* @param	name	登録する名前
	*/
	public static EventAdapter createAdapter( String name ) {
		EventAdapter ret = (EventAdapter )map.get( name );
		if ( ret != null ) {
			return ret;
		}
		ret = new EventAdapter();
		ret.name = name;
		map.put( name, ret );
		return ret;
	}

	/**
	* 指定された名前のEventAdapterを返します。
	* @param	name	登録されている名前
	*/
	public static EventAdapter getAdapter( String name ) {
		return (EventAdapter )map.get( name );
	}

	private class ServiceThread extends Thread {
		public void run() {
			try {
				for ( ;; ) {
					if ( isInterrupted() ) {
						break;
					}
					EventObject e = getEvent();
					if ( e == null ) {
						break;
					} else
					if ( e instanceof ReceiveEvent ) {
						fireReceive( (ReceiveEvent )e );
					} else
					if ( e instanceof StatusChangeEvent ) {
						fireStatusChange( (StatusChangeEvent )e );
					}
				}
			} catch ( Exception e ) {
				logger.error( this, "error", e );
			}
		}
		public synchronized void interrupt() {
			super.interrupt();
			notifyAll();
			super.interrupt();
		}
	}

	private EventObject getEvent() {
		EventObject e = null;
		for ( ;; ) {
			if ( isInterrupted()) {
				break;
			}
			try {
				if ( queue.isEmpty( 1000 )) {
					continue;
				}
			} catch ( InterruptedException ie ) {
				break;
			}
			e = (EventObject )queue.get();
			break;
		}
		return e;
	}

	/**
	* データに何らかの変更があったときに呼び出されます。
	* @param	e		イベントソースおよびデータの変更内容を指定します。
	*/
	public void received( ReceiveEvent e )
		throws Exception
	{
		if ( isDebug() ) {
			logger.debug( this, e.toString() );
		}
		queue.add( e );
	}

	/**
	* 状態に何らかの変更があったときに呼び出されます。
	* @param	e		イベントソースおよび状態の変更内容を指定します。
	*/
	public void statusChanged( StatusChangeEvent e )
		throws Exception
	{
		if ( isDebug() ) {
			logger.debug( this, e.toString() );
		}
		queue.add( e );
	}

	/**
	* データ受信を受け取るオブジェクトを追加します。
	* @param	l		データ受信を受け取るオブジェクト
	*/
	public void addReceiveListener( IReceiveListener l ) {
		support.addReceiveListener( (IReceiveListener)l );
	}
	/**
	* データ受信を受け取るオブジェクトを削除します。
	* @param	l		データ受信を受け取るオブジェクト
	*/
	public void removeReceiveListener( IReceiveListener l ) {
		support.removeReceiveListener( l );
	}
	/**
	* データ受信を受け取るオブジェクトを追加します。
	* @param	key		対象のキー
	* @param	l		データ受信を受け取るオブジェクト
	*/
	public void addReceiveListener( IDataKey key, IReceiveListener l ) {
		support.addReceiveListener( key, l );
	}
	/**
	* データ受信を受け取るオブジェクトを削除します。
	* @param	key		対象のキー
	* @param	l		データ受信を受け取るオブジェクト
	*/
	public void removeReceiveListener( IDataKey key, IReceiveListener l ){
		support.removeReceiveListener( key, l );
	}

	/**
	* データ受信イベントを通知します。
	* @param	e		データ受信イベント
	*/
	public void fireReceive( ReceiveEvent e )
		throws Exception
	{
		support.fireReceive( e );
	}

	/**
	* 状態変更イベントを通知します。
	* @param	e		状態変更イベント
	*/
	public void fireStatusChange( StatusChangeEvent e ) throws Exception {
		if ( isDebug()) {
			logger.debug( this, e.toString() );
		}
		support.fireStatusChange( e );
	}

	/**
	* この出力先の文字列表現を返します。
	* @return   この出力先の文字列表現
	*/
	public String toString() {
		String ret = super.toString();
		ret += ":name=[" + name + "]";
		return ret;
	}

}
